from .base.component import Component
from .base.enums import DataTypes
from .base.data_value import DataValue

class Solar(Component):
    def __init__(self) -> None:
        super().__init__(input_address=2100)
        self.collector_temperature_1 = DataValue(address=0,multiplier=0.1)
        self.collector_temperature_2 = DataValue(address=1,multiplier=0.1)
        self.collector_supply_temperature = DataValue(address=2,multiplier=0.1)
        self.collector_return_temperature = DataValue(address=3,multiplier=0.1)
        self.flow_heat_meter = DataValue(address=4,multiplier=0.1)
        self.curent_power = DataValue(address=5,multiplier=0.1)
        self.curent_yield_heat_meter = DataValue(address=6,count=2)
        self.today_yield = DataValue(address=8,count=2)
        self.buffer_sensor_1 = DataValue(address=10,multiplier=0.1)
        self.buffer_sensor_2 = DataValue(address=11,multiplier=0.1)
        self.buffer_sensor_3 = DataValue(address=12,multiplier=0.1)
        self.state = DataValue(address=13,type=DataTypes.UINT)
         


import logging
try:
    #modbus version < 3.0
    from pymodbus.client.sync import ModbusTcpClient as ModbusClient
    IS_LEGACY_VERSION = True
except:
    #modbus version >= 3.0
    from pymodbus.client import ModbusTcpClient as ModbusClient
    IS_LEGACY_VERSION = False
from .components.base.register_slice import RegisterSlice
  
class ModbusConnector():
    """
    Helper methodes to read/write data to a modbus server
    """
    def __init__(self,ip:str,port:int,slave_id:int) -> None:
        self.ip = ip
        self.port = port
        self.slave_id = slave_id
        self.client = ModbusClient(ip,port=port)
        self.__slave_args = {"unit":slave_id} if IS_LEGACY_VERSION else {"slave":slave_id}
        
    @property
    def is_connected(self)->bool:
        return self.client.is_socket_open()
    
    def connect(self)->bool:
        return self.client.connect()

  
    def read_input_registers(self,slices:list[RegisterSlice],count:int,check_connection:bool=True)->tuple[bool,list[int]]:
        """Internal methode to read input registers from modbus"""
        if check_connection and not self.is_connected:
            logging.error("Connection to modbus is not established!")
            return False, None
        try:
            combined_result = [None] * count
            for registerSlice in slices:
                result = self.client.read_input_registers(address=registerSlice.absolute_address,count=registerSlice.count, **self.__slave_args)
                if result.isError():
                    logging.error(f"Modbus read error at address={registerSlice.absolute_address}, count={registerSlice.count}: {result}")
                    return False, None
                slice_data = result.registers
                combined_result[registerSlice.relative_address:registerSlice.relative_address+registerSlice.count] = slice_data
            return True, combined_result
        except Exception:
            logging.exception(f"Exception while reading input registers for address: '{slices[0].absolute_address}'!")
            return False, None
     
    def read_holding_registers(self,slices:list[RegisterSlice],count:int, check_connection:bool = True)->tuple[bool,list[int]]:
        """Internal methode to read holding registers from modbus"""
        if check_connection and not self.is_connected:
            logging.error("Connection to modbus is not established!")
            return False, None
        try:
            combined_result = [None] * count
            for registerSlice in slices:
                result = self.client.read_holding_registers(address=registerSlice.absolute_address,count=registerSlice.count, **self.__slave_args)
                if result.isError():
                    logging.error(f"Modbus read error at address={registerSlice.absolute_address}: {result}")
                    return False, None
                slice_data = result.registers
                combined_result[registerSlice.relative_address:registerSlice.relative_address+registerSlice.count] = slice_data
            return True, combined_result
        except Exception:
            logging.exception(f"Exception while reading holding registers for address: '{slices[0].absolute_address}'!")
            return False, None
           
    def write_register(self,value:int, address:int, check_connection:bool = True) -> bool:
        """Write a value to the modbus server"""
        if check_connection and not self.is_connected:
            logging.error("Connection to modbus is not established!")
            return False
        try:
            response = self.client.write_registers(address, [value], **self.__slave_args)
            if response.isError():
                logging.error(f"Error writing value={value} to register: {address}: {response}")
                return False
        except Exception:
            logging.exception(f"Eception while writing value={value} to register: {address}!")
            return False
        return True
        
        
        


from abc import abstractmethod


class Part(object):
    """
    Abstraction of a metric of the heating system
    """
    
    @property
    @abstractmethod
    def scaled_value(self)->float:
        pass


"""Python client lib for Solarfocus"""
__version__ = "3.4.0"

from enum import Enum

#Default port for modbus
PORT = 502

#This needs to be defined before the imports because of circular dependencies
class Systems(str, Enum):
    """
    Supported systems by this library
    """
    Vampair = "Vampair"
    Therminator = "Therminator" 
    
from .modbus_wrapper import ModbusConnector
from .component_factory import ComponentFactory
from .const import SLAVE_ID

class SolarfocusAPI:
    """Solarfocus Heating System"""
    
    @property
    def system(self): 
        return self._system
    
    def __init__(self,
                 ip:str,
                 heating_circuit_count:int = 1,
                 buffer_count:int = 1,
                 boiler_count:int = 1,
                 system:Systems=Systems.Vampair,
                 port:int=PORT,
                 slave_id:int=SLAVE_ID):
        """Initialize Solarfocus communication."""
        assert heating_circuit_count >= 0 and heating_circuit_count < 9, "Heating circuit count must be between 0 and 8"
        assert buffer_count >= 0 and buffer_count < 5, "Buffer count must be between 0 and 4"
        assert boiler_count >= 0 and boiler_count < 5, "Boiler count must be between 0 and 4"
        
        self.__conn = ModbusConnector(ip,port,slave_id)
        self.__factory = ComponentFactory(self.__conn)
        #Lists of components
        self.heating_circuits = self.__factory.heating_circuit(system,heating_circuit_count)
        self.boilers = self.__factory.boiler(system,boiler_count)
        self.buffers = self.__factory.buffer(system,buffer_count)
        #Single components
        self.heatpump = self.__factory.heatpump(system)
        self.photovoltaic = self.__factory.photovoltaic(system)
        self.pelletsboiler = self.__factory.pelletsboiler(system)
        self.solar = self.__factory.solar(system)
        self._slave_id = slave_id
        self._system = system
    
    def connect(self):
        """Connect to Solarfocus eco manager-touch"""
        return self.__conn.connect()

    @property
    def is_connected(self)->bool:
        """Check if connection is established"""
        return self.__conn.is_connected
    
    def update(self):
        """Read values from Heating System"""
        if (
            self.update_heating()
            and self.update_buffer()
            and self.update_boiler()
            and self.update_heatpump()
            and self.update_photovoltaic()
            and self.update_pelletsboiler()
            and self.update_solar()
        ):
            return True
        return False

    def update_heating(self) -> bool:
        """Read values from Heating System"""
        for heating_circuit in self.heating_circuits:
            if not heating_circuit.update():
                return False
        return True
    
    def update_buffer(self) -> bool:
        """Read values from Heating System"""
        for buffer in self.buffers:
            if not buffer.update():
                return False
        return True
    
    def update_boiler(self) -> bool:
        """Read values from Heating System"""
        for boiler in self.boilers:
            if not boiler.update():
                return False
        return True

    def update_heatpump(self) -> bool:
        """Read values from Heating System"""
        return self.heatpump.update()

    def update_photovoltaic(self) -> bool:
        """Read values from Heating System"""
        return self.photovoltaic.update()

    def update_pelletsboiler(self) -> bool:
        """Read values from Pellets boiler"""
        return self.pelletsboiler.update()
    
    def update_solar(self) -> bool:
        """Read values from Solar"""
        return self.solar.update()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pysolarfocus', 'pysolarfocus.components', 'pysolarfocus.components.base']

package_data = \
{'': ['*']}

install_requires = \
['pymodbus==2.5.3']

setup_kwargs = {
    'name': 'pysolarfocus',
    'version': '3.4.0',
    'description': 'Unofficial, local Solarfocus client',
    'long_description': '# pysolarfocus: Python Client for Solarfocus eco<sup>_manager-touch_</sup>\n\n## What\'s Supported \n\n### Software Version\n\nThis integration has been tested with Solarfocus eco<sup>manager-touch</sup> version `21.040`.\n\nAdded biomass boiler pellet statistics (available since version 23.010)\n\n### Systems\n\n* Heat pump vamp<sup>air</sup> with eco<sup>_manager-touch_</sup>\n* Biomass boiler thermi<sup>nator</sup> II\n\n### Solarfocus Components\n\n| Components | Supported |\n|---|---|\n| Heating Circuits (_Heizkreis_) 1-8 | :white_check_mark: |\n| Buffers (_Puffer_) 1-4 | :white_check_mark: |\n| Solar (_Solar_)| :white_check_mark:|\n| Boilers (_Boiler_) 1-4 | :white_check_mark: |\n| Heatpump (_Wärmepumpe_) | :white_check_mark: |\n| Biomassboiler (_Kessel_) | :white_check_mark: | \n\n## Usage\n\n```python\nfrom pysolarfocus import SolarfocusAPI,Systems\n\n# Create the Solarfocus API client\nsolarfocus = SolarfocusAPI(ip="[Your-IP]",system=Systems.Vampair)\n# Connect to the heating system\nsolarfocus.connect() \n# Fetch the values\nsolarfocus.update()\n\n# Print the values\nprint(solarfocus.buffers[0])\nprint(solarfocus.heating_circuit[0])\n```\n\n### Handling multiple components e.g. heating circuits\nSolarfocus systems allow the use of multiple heating circuits, buffers and boilers. The api can be configured to interact with multiple components.\n\n```python \n\n# Create the Solarfocus API client with 2 Heating Circuits\nsolarfocus = SolarfocusAPI(ip="[Your-IP]",heating_circuit_count=2,system=Systems.Vampair)\n# Connect to the heating system\nsolarfocus.connect()\n\n# Update all heating circuits\nsolarfocus.update_heating()\n\n# Update only the first heating circuit\nsolarfocus.heating_circuits[0].update()\n# Print the first heating circuit\nprint(solarfocus.heating_circuits[0])\n\n# Set the temperature of the first heating circuit to 30°C\nsolarfocus.heating_circuits[0].indoor_temperatur_external.set_unscaled_value(30)\n# Write the value to the heating system\nsolarfocus.heating_circuits[0].indoor_temperatur_external.commit()',
    'author': 'Jeroen Laverman',
    'author_email': 'jjlaverman@web.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/lavermanjj/pysolarfocus',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from .part import Part

class PerformanceCalculator(Part):
    """
    Performing performance calculations
    """
    
    nominator:Part
    denominator:Part
    value:int = None
    
    def __init__(self, nominator:Part, denominator:Part) -> None:
        self.nominator = nominator
        self.denominator = denominator
    
    @property
    def value(self)->float:
        """
        Return value of the calculation
        """
        if self.denominator.scaled_value:
            return self.nominator.scaled_value / self.denominator.scaled_value
        return 0.0
    
    @property
    def scaled_value(self)->float:
        """
        To provide a uniform interface for integrating, a scaled value property is offered
        """
        return self.value
            


from .base.component import Component
from .base.enums import DataTypes, RegisterTypes
from .base.data_value import DataValue

class PelletsBoiler(Component):
    def __init__(self,input_address=2400,holding_address=33400) -> None:
        super().__init__(self,input_address, holding_address)
        self.temperature = DataValue(address=0,multiplier=0.1)
        self.status  = DataValue(address=1,type=DataTypes.UINT)
        self.message_number = DataValue(address=4)
        self.door_contact = DataValue(address=5)
        self.cleaning = DataValue(address=6)
        self.ash_container = DataValue(address=7)
        self.outdoor_temperature = DataValue(address=8,multiplier=0.1)
        self.boiler_operating_mode = DataValue(address=9)
        self.octoplus_buffer_temperature_bottom = DataValue(address=10,multiplier=0.1)
        self.octoplus_buffer_temperature_top = DataValue(address=11,multiplier=0.1)
        self.log_wood = DataValue(address=12,type=DataTypes.UINT)
        self.pellet_usage_last_fill = DataValue(address=14,type=DataTypes.UINT,multiplier=0.1)
        self.pellet_usage_total = DataValue(address=16,type=DataTypes.UINT,multiplier=0.1)
        self.heat_energy_total = DataValue(address=18,type=DataTypes.UINT,multiplier=0.1)
        self.pellet_usage_reset = DataValue(address=12,type=DataTypes.UINT,register_type=RegisterTypes.Holding)


from .base.component import Component
from .base.enums import DataTypes,RegisterTypes
from .base.data_value import DataValue

class HeatingCircuit(Component):
    def __init__(self,input_address=1100,holding_address=32600) -> None:
        super().__init__(input_address=input_address,holding_address=holding_address)
        self.supply_temperature = DataValue(address=0,multiplier=0.1)
        self.room_temperature = DataValue(address=1,multiplier=0.1)
        self.humidity = DataValue(address=2,multiplier=0.1)
        self.limit_thermostat = DataValue(address=3,type=DataTypes.UINT)
        self.circulator_pump = DataValue(address=4,type=DataTypes.UINT)
        self.mixer_valve = DataValue(address=5,type=DataTypes.UINT)
        self.state = DataValue(address=6,type=DataTypes.UINT)
        
        self.target_supply_temperature = DataValue(address=0,multiplier=10,register_type=RegisterTypes.Holding)
        self.cooling = DataValue(address=2,register_type=RegisterTypes.Holding)
        self.mode = DataValue(address=3,register_type=RegisterTypes.Holding)
        self.target_room_temperatur = DataValue(address=5,multiplier=10,register_type=RegisterTypes.Holding)
        self.indoor_temperatur_external = DataValue(address=6,multiplier=10,register_type=RegisterTypes.Holding)
        self.indoor_humidity_external = DataValue(address=7,register_type=RegisterTypes.Holding)
        
class TherminatorHeatingCircuit(HeatingCircuit):
    def __init__(self, input_address=1100, holding_address=32600) -> None:
        super().__init__(input_address, holding_address)
        # No idea, why this is offset by 1 ... 
        self.circulator_pump = DataValue(address=5,type=DataTypes.UINT)
        self.mixer_valve = DataValue(address=6,type=DataTypes.UINT)
        self.state = DataValue(address=7,type=DataTypes.UINT)